/*
  For Lilygo T-Display-S3-Long 
  nikthefix - 28th Nov 2023
  Modified driver code and example sketch using TFT_eSPI in 'sprite only mode' 

  Versions:
  TFT_eSPI 2.5.34 - latest at time of writing
  ESP32 Arduino 3.0.0-alpha2 - latest at time of writing

  Notes:
  As the display doesn't implement a scan orientation hardware rotate - as far as I can see from the current datasheet - we need to use 
  a soft matrix rotation to get a landscape view without messing with TFT_eSPI. This is implemented in lcd_PushColors_rotated_90().
  You can then use lcd_setRotation(2) which IS hardware implemented, to flip the whole thing upside down if you need.
  Code has been stripped down to support QSPI display only.

  Build Options:
  Board  ESP32-S3-Dev
  USB CDC On oot Enabled
  Flash Size 16MB
  Partition Scheme 16M Flash(3MB APP/9.9MB FATFS)
  PSRAM "OPI PSRAM"

  Since ESP32 Arduino 3.0.0-alpha2 is still pretty funky with a lot of existing Arduino driver code it may be necessary to downgrade to V2.xx as the project expands - in the short term
*/

#include "AXS15231B.h"
#include <Arduino.h>

#include <Wire.h>
#include "pins_config.h"


/**
 * WiFiManager advanced demo, contains advanced configurartion options
 * Implements TRIGGEN_PIN button press, press for ondemand configportal, hold for 3 seconds for reset settings.
 */
#include <WiFiManager.h> // https://github.com/tzapu/WiFiManager

// wifimanager can run in a blocking mode or a non blocking mode
// Be sure to know how to process loops with no delay() if using non blocking
bool wm_nonblocking = false; // change to true to use non blocking

WiFiManager wm; // global wm instance
WiFiManagerParameter custom_field; // global param ( for non blocking w params )

//Over the air updates
#include <AsyncElegantOTA.h>
const char *ssid = "........";
const char *password = "........";
bool updateMode = 0;           // pending OTA update mode
String otaIP;
AsyncWebServer webServer(80);

//UDP
#include <WiFi.h>
#include <WiFiClient.h>
WiFiUDP udp;

const uint16_t port  = 49155;
#define WIFIPOWER WIFI_POWER_MINUS_1dBm
const uint16_t  LOOPTIME          = 125;    			// milliseconds.
const uint16_t  FRAMES            = 2000 / LOOPTIME;    // number of frames before link failure detected
const bool DEBUG = false;
uint64_t currentMillis = millis();
uint64_t previousMillis = millis();
int16_t  frameCount = FRAMES;  // number of sequential frames missing before display is disabled


#include "dynon.h"
#include "parseDynon.h"
EFISdata myEFISdata;

//Preferences and version/serial data
#include "userData.h" //holds device's serial number and aircraft settings
#include "version.h"

#include "LIDARLite_v4LED.h"
LIDARLite_v4LED myLIDAR;
uint8_t myLIDAR_address = 0x62;
bool hasLidar = false;
float newDistance;
int Distance;
float Feet;
float mmToFeet;
int sFeet; //Converted to feet height from the ground
//int sensorFt = 2; //sensor is mounted 2ft from the ground
int previousFeet = 0;

#include "SparkFun_STTS22H.h"
SparkFun_STTS22H sensor0;
uint8_t STTS22H_address = 0x3C;
float temperature;
bool hasTemp=false; //true if temp sensor is available

//Control vars
bool isFlying=false;
bool showLidar=false;
bool hasDynon=false;
uint useColor; // use to test current data against parameters
int curScreen = 1; //0-OTA/setup, 1-CHECKLIST, 2-EMS, 3-EFIS, 4-SETTINGS
int dX; //used to help build X positions
int dY; //used to help build Y positions
int hY; //used to help build height positions in EGT/CHT chart animations
float conv; //used to convert gallons (int to float)
float cgtDif; //used to calculate EGT/CHT chart animations

bool res = false; //check if network is connected


#include <TFT_eSPI.h>
// For bitmap encoding: use Image2lcd, 16bit true colour, MSB First, RGB565, don't include head data, be sure to set max image size, save as .h file.

TFT_eSPI tft = TFT_eSPI();
TFT_eSprite sprite = TFT_eSprite(&tft);
TFT_eSprite back = TFT_eSprite(&tft);
TFT_eSprite sky = TFT_eSprite(&tft);
#include "font.h"

//Colors and fonts
#define cGreen 0x66F0
#define cBlue 0x86bf //light blue
#define cDarkBlue 0x1dda //dark blue
#define cYellow 0xfe20 //yellow background 0xFE20
#define cRed 0xC0E3 //red background
#define cGray 0x7BCF //gray background 0x7bcf
#define cDarkGray 0x18a3 //gray buttons
#define cIconGray 0x528a //gray icons
#define cPink 0xFA1F 

#define cWhite TFT_WHITE
#define color1 TFT_WHITE

#define digit30 DS_DIGI30pt7b 
#define digit24 DS_DIGI24pt7b 
#define digit18 DS_DIGI18pt7b
#define digit16 DS_DIGI16pt7b
#define digit14 DS_DIGI14pt7b
#define digit12 DS_DIGI12pt7b
#define din7 DIN_Condensed_Bold7pt7b
#define din8 DIN_Condensed_Bold8pt7b
#define din12 DIN_Condensed_Bold12pt7b
#define din16 DIN_Condensed_Bold16pt7b
#define din18 DIN_Condensed_Bold18pt7b
#define din24 DIN_Condensed_Bold24pt7b
const uint16_t SCREEN_X = 640;  // Screen sizes are set here.
const uint16_t SCREEN_Y = 180;
const char *prodName = "CoFlier_Pro";
int pHead = 0; //previous heading

//Checklists
int curCheck = 0; //start with first checklist
int curItem = 0; //start with first item of the checklist
int totItem = totPreflight; //starts with first checklist
int lastCheck = 0; //used to avoid double tap
String taskMsg = ""; //used with checklists

#include "setupScreens.h"
#include "homeScreen.h"
#include "emsScreen.h"
#include "efisScreen.h"

//Screen touch setup
uint8_t ALS_ADDRESS = 0x3B;
uint8_t read_touchpad_cmd[11] = {0xb5, 0xab, 0xa5, 0x5a, 0x0, 0x0, 0x0, 0x8};
bool touch_held=false;
#define time_out_reset 30000
int touch_timeout=time_out_reset;
int pointX=-1;
int pointY=-1;
int type = 0;
   
void setup() {
  pinMode(TOUCH_INT, INPUT_PULLUP);

  // comment out if using variable brightness
  pinMode(TFT_BL, OUTPUT);
  digitalWrite(TFT_BL, HIGH);
  Serial.begin(115200);

  // if using esp32 3.0.0-alpha3 then you can use the following 2 lines to adjust the backlight brightness
  //ledcAttach(TFT_BL, 10000, 8);// pin, pwm freq, resolution in bits
  //ledcWrite(TFT_BL, 30);// pin, pulse width (a.k.a brightness 0-255)

  pinMode(TOUCH_RES, OUTPUT);
  digitalWrite(TOUCH_RES, HIGH);delay(2);
  digitalWrite(TOUCH_RES, LOW);delay(10);
  digitalWrite(TOUCH_RES, HIGH);delay(2);
  Wire.begin(TOUCH_IICSDA, TOUCH_IICSCL);

  axs15231_init();
  Wire1.begin(43,44);
  
  if(!sensor0.begin(STTS22H_address))
  {
    Serial.println("Cannot connect to STTS22H.");
    //while(1);
  } else { Serial.println("Connected to STTS22H.");
    hasTemp = true;
    sensor0.setDataRate(STTS22H_POWER_DOWN);
    delay(10);
    sensor0.setDataRate(STTS22H_25Hz);
    sensor0.enableAutoIncrement();
    sensor0.getTemperatureF(&temperature);
    Serial.println(temperature);
  }


  if(!myLIDAR.begin(myLIDAR_address, Wire1))
  {
    Serial.println("Cannot connect to Lidar.");
    //while(1);
  } else { 
    Serial.println("Connected to Lidar.");
    hasLidar = true;   
  }

  curScreen = 1;
  bool res;
  res = 0;
  //axs15231_init();

  lcd_setRotation(2);             // 180 degree hardware rotate if you want reset / boot buttons at the bottom
  //sprite.setColorDepth(16);
  sprite.createSprite(SCREEN_X,SCREEN_Y);    // full screen landscape sprite in psram
  sprite.setSwapBytes(1);
  //back.createSprite(SCREEN_X,SCREEN_Y);
  sky.createSprite(SCREEN_X*2,SCREEN_Y*2);
  sprite.setPivot(SCREEN_X/2,SCREEN_Y/2);
  lcd_fill(0,0,180,640,0x00);       // clear screen
  digitalWrite(TFT_BL, HIGH);       // turn on backlight

  drawMaster();

  WiFi.mode(WIFI_STA);
  if(wm_nonblocking) wm.setConfigPortalBlocking(false);

  // **REVIEW IN THE FUTURE ** Add a custom input field
  int customFieldLength = 40;
  const char* custom_radio_str = "<br/><label for='customfieldid'>Custom Field Label</label><input type='radio' name='customfieldid' value='1' checked> One<br><input type='radio' name='customfieldid' value='2'> Two<br><input type='radio' name='customfieldid' value='3'> Three";
  new (&custom_field) WiFiManagerParameter(custom_radio_str); // custom html input 
  wm.addParameter(&custom_field);
  wm.setSaveParamsCallback(saveParamCallback);
  std::vector<const char *> menu = {"wifi","info","param","sep","restart","exit"};
  wm.setMenu(menu);

  // set dark theme
  wm.setClass("invert");
  wm.setConfigPortalTimeout(30); // auto close configportal after n seconds
  res = wm.autoConnect(prodName,""); // password protected ap

  if(!res) {
    Serial.println("Failed to connect or hit timeout");
    failedConnect();
    // ESP.restart();
  } 
  else {
    //if you get here you have connected to the WiFi    
    Serial.println("connected...yeey :)");
    Serial.print(F("Connected. Local ip: ")); 
    Serial.println(WiFi.localIP());
    udp.begin(port);
    Serial.println ("WiFi Station Connected, listening for UDP Broadcasts");
    Serial.print ("Local IP address: "); Serial.println (WiFi.localIP());
    Serial.print("Listening on UDP port ");
    Serial.println(port);
    //startingDevice();
    curScreen = 1; //activates HOMESCREEN - checklist
  }

  if (updateMode) {
    otaIP = WiFi.localIP().toString();
    //OTA
    webServer.on("/", HTTP_GET, [](AsyncWebServerRequest *request) {
      request->send(200, "text/plain", "CoPilot Pro Updater.");
    });

    AsyncElegantOTA.begin(&webServer);    // Start AsyncElegantOTA
    webServer.begin();
    Serial.println("HTTP server started");
    drawOTA();
  }

    
  //drawMaster();

  Serial.println(F("setup(): finished"));
}

void loop() {     
  if (updateMode == false) {
    // Reads Dynon data
    long unsigned int time1 = micros();
    static int16_t msgTime = 0;
    static const char *myMessage = "20 character message";

    AsyncElegantOTA.loop();

    currentMillis = millis();
    if ((currentMillis - previousMillis) >= 5000) {
      if (WiFi.status() != WL_CONNECTED) {
        Serial.printf("WiFi disconnected, reconnecting...\n");
        WiFi.disconnect();
        WiFi.mode(WIFI_AP_STA);
        WiFi.begin(ssid, password);
        WiFi.setTxPower(WIFIPOWER);
      }
      previousMillis = currentMillis;
    }
    int packetSize = udp.parsePacket();
    //Serial.print ("PacketSize: "); Serial.println (packetSize);
    if (packetSize > 0){
      hasDynon = true;
      // read the packet into packetChars
      Serial.printf("Received packet of size %d ", packetSize);
      IPAddress remoteIp = udp.remoteIP();
      Serial.printf("From Port %d ", udp.remotePort());

      int len = udp.read(myEFISdata.packetChars, NUMCHARS);
      if (len > 0) {
        myEFISdata.packetChars[len] = 0;
      }
      // Extract the data fields.
      Serial.printf ("Contents: %s \n", myEFISdata.packetChars);
      myEFISdata.udpPacketFrame(myEFISdata.packetChars, len);
    }
    if (myEFISdata.packetValid) {  // a sentence has been detected.
      frameCount = FRAMES;  // indicates serial link activity.
    }
    myEFISdata.packetValid = false; // release framed packet

    //Reads Lidar
    if (hasLidar) {
      mmToFeet = 0.032808399; //1 mm = this in feet
      newDistance = myLIDAR.getDistance();
      Distance = newDistance;
      Feet = newDistance * mmToFeet - sensorFt; 
      if (Feet < 0)
      {
        Feet = 0;   // Don't go negative due to offset
      }
      sFeet = Feet;
    }

    if(digitalRead(TOUCH_INT)==0) {//} && !touch_held) {
      getTouch();
        //drawMaster();        
      switch (curScreen) { 
        case 1:
          drawMaster();
          break;
        case 2:
          drawEMS();
          break;
        case 3:
          if (pHead != Heading){
            if(Heading>pHead){
              if (Heading>359 || Heading <0) { Heading = 0;}
            } else {
              if (Heading>359 || Heading <0) { Heading = 360;}
            }
            pHead = Heading;
          } 
          //Serial.println("Heading: "+String(Heading)+" pHead: "+String(pHead));
          drawEFIS();
          break;
        case 4:
          configNetwork();
          break;
        default:
          //drawEMS();
          break;
      }
      
      touch_held=true;
      touch_timeout=time_out_reset;
    } else {
        touch_timeout--;
        if(touch_timeout == 0) {
          touch_held=false;
          touch_timeout=time_out_reset;
        } 
    } 
   } //end if updatemode 
  
}

void nextChecklist(uint8_t eventFire) {
  Serial.println("Event fire: "+String(eventFire));
  Serial.println("curCheck: "+String(curCheck)+" lastCheck: "+String(lastCheck));
  if (eventFire==1) { //Emergency request
    curCheck = 12;
    Serial.println("First time Emergency: "+String(curCheck));
    lastCheck=curCheck;
  } else if (eventFire==2 && curCheck!=lastCheck) { //Next Emergency checklist request
    Serial.println("IN event2 DIFFERENT lastCheck: "+String(curCheck));
    lastCheck=curCheck;
  } else if (eventFire==2 && curCheck==lastCheck) { //Next Emergency checklist request
    curCheck = curCheck + 1;
    Serial.println("IN event2 SAME lastCheck: "+String(curCheck));
    if (curCheck > totChecklists-1) { curCheck = 12;}
    lastCheck=curCheck;
  } else if (eventFire==0 && curCheck==lastCheck) {
    curCheck = curCheck + 1;
    Serial.println("REGULAR Go to checklist: "+String(curCheck));
    if (curCheck > totChecklists-9) { curCheck = 0;}
    lastCheck=curCheck;
  } else {
    lastCheck=curCheck;
  }
  Serial.println("Go to checklist: "+String(curCheck));
  
  switch (curCheck) {
    case 0:
      totItem = totPreflight;
      break;
    case 1:
        totItem = totStartCold;
        break;
    case 2:
        totItem = totStartHot;
        break;
    case 3:
        totItem = totBeforeTaxi;
        break;
    case 4:
        totItem = totRunup;
        break;
    case 5:
        totItem = totBeforeTakeoff;
        break;
    case 6:
        totItem = totAfterTakeoff;
        break;
    case 7:
        totItem = totCruise;
        break;
    case 8:
        totItem = totDescent;
        break;
    case 9:
        totItem = totPattern;
        break;
    case 10:
        totItem = totAfterLanding;
        break;
    case 11:
        totItem = totShutdown;
        break;
    case 12:
        totItem = totEmergency1;
        break;
    case 13:
        totItem = totEmergency2;
        break;
    case 14:
        totItem = totEmergency3;
        break;
    case 15:
        totItem = totEmergency4;
        break;
    case 16:
        totItem = totEmergency5;
        break;
    case 17:
        totItem = totEmergency6;
        break;
    case 18:
        totItem = totEmergency7;
        break;
    case 19:
        totItem = totEmergency8;
        break;
  }
  curItem = 0; //Uses 1 because the double tap returns to the previous checklist when using 0
}


void newCheckItem() { 
  if (pointX >= SCREEN_X/2) {
    //Scenario 1: Advance task in the current checklist when touch is right side of the screen
    Serial.println("NEXT task");
    curItem = curItem + 1;
    if (curItem > totItem-1) { 
      curItem = 0;
      curCheck = curCheck + 1;
      if (curCheck > totChecklists-1) {  //Emergency checklists
        curCheck = 12;
      } else if (curCheck > totChecklists-9) { //Regular checklists
        curCheck = 0;
      }
    }
  } else { 
    //Scenario 2: Previous task in the current checklist when touch is left side of the screen
    Serial.println("PREVIOUS task");
    curItem = curItem - 1;
    if (curItem<0) { //Need to return to the previous checklist
      curCheck = curCheck - 1;
      if (curCheck<0) { 
        curCheck= totChecklists-9; 
      } else if (curCheck==11) {
        curCheck = totChecklists-1; //returns to the first emergency in the case user is already in the EMERGENCY situation
      } 
      switch (curCheck) {
        case 0:
          totItem = totPreflight;
          break;
        case 1:
            totItem = totStartCold;
            break;
        case 2:
            totItem = totStartHot;
            break;
        case 3:
            totItem = totBeforeTaxi;
            break;
        case 4:
            totItem = totRunup;
            break;
        case 5:
            totItem = totBeforeTakeoff;
            break;
        case 6:
            totItem = totAfterTakeoff;
            break;
        case 7:
            totItem = totCruise;
            break;
        case 8:
            totItem = totDescent;
            break;
        case 9:
            totItem = totPattern;
            break;
        case 10:
            totItem = totAfterLanding;
            break;
        case 11:
            totItem = totShutdown;
            break;
        case 12:
            totItem = totEmergency1;
            break;
        case 13:
            totItem = totEmergency2;
            break;
        case 14:
            totItem = totEmergency3;
            break;
        case 15:
            totItem = totEmergency4;
            break;
        case 16:
            totItem = totEmergency5;
            break;
        case 17:
            totItem = totEmergency6;
            break;
        case 18:
            totItem = totEmergency7;
            break;
        case 19:
            totItem = totEmergency8;
            break;
      }
      curItem = totItem-1;
    }
  }
}

void getTouch() {
    //Serial.print("TOUCHED :: touch_held: ");
    //Serial.println(touch_held);
    if(touch_held==true) return;

    uint8_t buff[20] = {0};
    Wire.beginTransmission(ALS_ADDRESS);
    Wire.write(read_touchpad_cmd, 8);
    Wire.endTransmission();
    Wire.requestFrom(ALS_ADDRESS, 8);
    while (!Wire.available());
    Wire.readBytes(buff, 8);

    type = AXS_GET_GESTURE_TYPE(buff);
    pointX = AXS_GET_POINT_X(buff,0);
    pointY = AXS_GET_POINT_Y(buff,0);

    Serial.println(pointX);
    if(pointX > 640) pointX = 640;
    if(pointY > 180) pointY = 180;

    //Buttons
    //SCENARIOS FOR CHECKLIST SCREEN
    if (curScreen==1) {
      //Scenario 0: touched the bottom buttons
      Serial.print("PointX: ");
      Serial.print(pointX);
      Serial.print(" PointY: ");
      Serial.println(pointY);
      if (pointY>120) {
        //work with the X position for each button ** pointX is inverted due to the TFT_eSPI implementation
        if (pointX > 528) {
          Serial.println("Tap button 5 - SETTINGS");  
          curScreen=4;    
        } else if (pointX > 415) {          
          Serial.println("Tap button 4 - EFIS");
          curScreen=3;
        } else if (pointX > 226) {
          Serial.println("Tap button 3 - EMERGENCY");
          if (curCheck<12) {
            nextChecklist(1); //FIRST EMERGENCY
          } else {
            nextChecklist(2);
          }
        } else if (pointX > 113) {
          Serial.println("Tap button 2 - EMS");
          curScreen=2;
        } else {
          Serial.println("Tap button 1 - Go to next CHECKLIST");
          nextChecklist(0);
        }
      } else {
        newCheckItem();
      }
    } else {
      //From other screens to the main (homeScreen)
      Serial.println("Back HOME");
      curScreen=1;
      drawMaster();
    }

   
}

//USED IN NETWORK SETUP OPTIONAL PARAMETERS - maybe for engine and imperial/metric selection
String getParam(String name){
  //read parameter from server, for customhmtl input
  String value;
  if(wm.server->hasArg(name)) {
    value = wm.server->arg(name);
  }
  return value;
}
void saveParamCallback(){
  Serial.println("[CALLBACK] saveParamCallback fired");
  Serial.println("PARAM customfieldid = " + getParam("customfieldid"));
}